#include "WiiPCCommon.h"
#include "../../Core/Src/HW/Wiimote.h"

#include "WiiPCReader.h"
#include "WiiPCWriter.h"
#include "WiiPCMapper.h"
#include "WiiPCPresenter.h"
#include "WiiPCEventer.h"
#include "WiiPCManager.h"
#include "dconsole.h"

extern int Index_0_Header;
extern int Index_1_Header;
extern int Wii_Balance_Board_Flag;

namespace WiiPC {

uint16_t    current_connection;
uint8_t     bridge_mode   = 0;
uint8_t     _connected_num = 0;
uint8_t     _connected_first = 0;

ReaderT*    _Reader    = 0;
WriterT*    _Writer    = 0;
ManagerT*   _Manager   = 0;
PresenterT* _Presenter = 0;
EventerT*   _Eventer   = 0;
MapperT*    _Mapper    = 0;

/* Transx: transactions */
std::mutex            mutex_transx_vars;
std::recursive_mutex  mutex_transx;

uint8_t               transx_major;
uint8_t               transx_minor;
bool                  transx_waiting;
bool                  transx_received;
udata_t               transx_udata;

ReaderT*    Reader(void){
    if(_Reader == 0) {
        _Reader = new ReaderT();
    }

    return _Reader;
}

WriterT*    Writer(void){
    if(_Writer == 0) {
        _Writer = new WriterT();
    }

    return _Writer;
}

ManagerT*   Manager(void){
    if(_Manager == 0) {
        _Manager = new ManagerT();
    }

    return _Manager;
}

PresenterT* Presenter(void){
    if(_Presenter == 0) {
        _Presenter = new PresenterT();
    }

    return _Presenter;
}

EventerT*   Eventer(void){
    if(_Eventer == 0) {
        _Eventer = new EventerT();
    }

    return _Eventer;
}

MapperT*    Mapper(void){
    if(_Mapper == 0){
        _Mapper = new MapperT();
    }

    return _Mapper;
}

uparam_t*   SetupUevent(uevent_t* e, uint8_t major, uint8_t minor, 
                        uint16_t num_params, uint8_t len_tail){
    if(e) {
        e->major = major;
        e->minor = minor;
        e->num_params = (uint8_t)num_params;
        e->param->u.head.len_tail = len_tail;
        e->param->u.head.num_params = (uint8_t)num_params;

        return e->param;
    }

    return 0;
}

uparam_t* SetupUdataUevent(udata_t* d, uint8_t major, 
                           uint8_t minor, uint8_t nparams, uint8_t nb_tail){
    d->type = udata_type_uevent;
    d->head.usb_report_id = USB_EXTRA_RX_REPORT_ID;
    d->head.cmd = udata_usb_extra_cmd_uevent;
    d->uevent.body.major = major;
    d->uevent.body.minor = minor;
    d->uevent.body.num_params = nparams;
    d->uevent.body.param[0].u.head.len_tail = nb_tail;
    d->uevent.body.param[0].u.head.num_params = nparams;	

    return d->uevent.body.param;
}

uint16_t UeventType(uevent_t* uevent){
    return (((uint16_t)uevent->major) << 8) | (uevent->minor & 0xFF);
}

/*
    Shutdown Agent
    To make sure Shutdown() is called while the program is ended.
*/
struct AgentT {
    ~AgentT(){
        Shutdown();
    }
} _agent;


uparam_t* Transx(uevent_t* uevent,
                 uint8_t cfm_major, 
                 uint8_t cfm_minor, 
                 uint8_t cmd = 0,
                 bool write_only = false,
                 uint32_t* operation_result = 0,
                 uint32_t timeout_msec = 0)
{
    std::lock_guard<std::recursive_mutex> lk(mutex_transx);
    const uint32_t interval = 10 /* ms */;

	// default_timeout_msec setting:		
	// Pary set 1.0 sec
	// Webb set 1.5 sec
    const uint32_t default_timeout_msec = 1500; /* 1.5 sec */	
    uint32_t timeout_count;
    uint32_t result;
    udata_t udata = {0};	

    if(ManagerIsUsbDisconnected()){
        if(operation_result){
            *operation_result = ERR_DEVICE_ABSENT;
        }

        return 0;
    }

    if(cmd) {		
        SetupUdataUevent(&udata, 0, 0, 0, 0);
        udata.uevent.cmd = cmd;
    }else{		
        SetupUdataUevent(&udata,
                         uevent->major,
                         uevent->minor,
                         uevent->param[0].u.head.num_params,
                         uevent->param[0].u.head.len_tail);
    }	

    hton_uparam(uevent->param, udata.uevent_raw.raw);

    mutex_transx_vars.lock();
    transx_major = cfm_major;
    transx_minor = cfm_minor;
    transx_received = false;
    transx_waiting  = write_only ? false : true;
	
    result = WriterWrite(&udata);
    mutex_transx_vars.unlock();

    if(operation_result){
        *operation_result = result;
    }

    if(write_only){
        return 0;
    }

    timeout_count = timeout_msec ? 
        timeout_msec / interval : 
        default_timeout_msec / interval /* = 1 sec */;

    while(!transx_received && timeout_count) {
        ReaderTryRead(30);
        Sleep(interval);
        timeout_count --;
    }

    mutex_transx_vars.lock();
    transx_waiting = false;
    mutex_transx_vars.unlock();

    if(timeout_count == 0){
        if(operation_result) {
            *operation_result = ERR_TIMEOUT;
        }
    }

    return timeout_count ? transx_udata.uevent.body.param : 0;
}

#define TRANS_SENDX_MSEC(MAJOR, MINOR, NUM, TAIL, MSEC) \
    uevent_t uevent; uparam_t* V;  bool OK = 0; uint32_t __Result; int __S0 = 0; uparam_t* U; \
    for(__S0=0, U = SetupUevent(&uevent, umajor_##MAJOR, MAJOR##_minor_cmd_##MINOR, NUM, TAIL); \
        bridge_mode && __S0==0; \
        __S0++, U = U, V = Transx(&uevent, umajor_##MAJOR, MAJOR##_minor_cfm_##MINOR, 0, false, &__Result, MSEC), OK = (__Result == ERR_OK) )

#define TRANS_SENDX(MAJOR, MINOR, NUM, TAIL)    TRANS_SENDX_MSEC(MAJOR, MINOR, NUM, TAIL, 1000)


bool query_status(uint8_t* bridge_mode, uint8_t* conn_bitmap, uint8_t* ir_status, uint8_t* ir_pos, uint8_t* discovering){
    uevent_t uevent;

    SetupUevent(&uevent, 0, 0, 0, 0);
    uparam_t* v = Transx(&uevent, umajor_ucore, ucore_minor_cfm_status, udata_usb_extra_cmd_status);	

    if(v){
        if(bridge_mode){
            *bridge_mode       = v[1].u.u8;
        }

        if(conn_bitmap){
            uint8_t i;
            uint8_t bitmap = v[2].u.u8;

            *conn_bitmap = bitmap;

            /* update connected number */
            _connected_num = 0;

            for(i=0; i<4; ++i) {
                if(bitmap & (1<<i)) {
                    _connected_num ++;
                }
            }

            for(i=0; i<4; ++i) {
                if(bitmap & (1<<i)) {
                    _connected_first = i;
                    break;
                }
            }
        }

        if(ir_status){
            *ir_status         = v[3].u.u8;
        }

        if(ir_pos){
            *ir_pos            = v[4].u.u8;
        }

        if(discovering) {
            *discovering       = v[5].u.u8;
        }

        return true;
    }

    return false;
}

bool enable_bridge(void){
    uint32_t result;
    uevent_t uevent;
    SetupUevent(&uevent, 0, 0, 0, 0);
    Transx(&uevent, 0, 0, udata_usb_extra_cmd_enable_bridge, true /* write_only */, &result);

    if(result){
        return false; /* NG */
    }

    return true; /* OK */
}

bool disable_bridge(void){
    uint32_t result;
    uevent_t uevent;
    SetupUevent(&uevent, 0, 0, 0, 0);
    Transx(&uevent, 0, 0, udata_usb_extra_cmd_disable_bridge, true /* write_only */, &result);
    return result == ERR_OK;
}

bool send_present(void){
    uint32_t result;
    uevent_t uevent;
    SetupUevent(&uevent, 0, 0, 0, 0);
    Transx(&uevent, 0, 0, udata_usb_extra_cmd_present, true /* write_only */, &result);
    return result == ERR_OK;
}

bool send_absent(void){
    uint32_t result;
    uevent_t uevent;
    SetupUevent(&uevent, 0, 0, 0, 0);
    Transx(&uevent, 0, 0, udata_usb_extra_cmd_absent, true /* write_only */, &result);
    return result == ERR_OK;
}

bool wiimote_get_states(uint8_t states[4]) {
    uint8_t i;

    TRANS_SENDX(hidhost, get_states, 0, 0){
    }

    /* set default values */
    for(i=0; i<4; i++){
        states[i] = 0;
    }

    if(OK){
        for(i=0; i<4; i++){
            states[i] = V[i+1].u.u8;
        }
    }

    return OK;
}

bool wiimote_disconnect(uint8_t device_index) {
    uint32_t result;
    uevent_t uevent;
    uparam_t* u = SetupUevent(&uevent, 
                              umajor_hidhost, 
                              hidhost_minor_cmd_disconnect, 
                              1 /* num_params */, 
                              0 /* len_tail */);
    uparam_set(u[1], u8, device_index);
    Transx(&uevent, 0, 0, 0, true /* write_only */, &result);
    return result == ERR_OK;
}

bool wiimote_set_rumble(uint8_t index, uint8_t enabled) {
    uint32_t result;
    uevent_t uevent;
    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_set_rumble, 2, 0);
    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, enabled);
    Transx(&uevent, 0, 0, 0, true /* write_only */, &result);

    return result == ERR_OK;
}

bool wiimote_set_leds(uint8_t index, uint8_t leds, uint8_t ack) {
    uint32_t result;
    uevent_t uevent;
    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_set_leds, 3, 0);

    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, leds);
    uparam_set(u[3], u8, ack);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_set_channel(uint8_t index, uint8_t channel, uint8_t continuous, uint8_t ack) {
    uint32_t result;
    uevent_t uevent;
    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_set_channel, 4, 0);

    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, channel);
    uparam_set(u[3], u8, continuous);
    uparam_set(u[4], u8, ack);

    Transx(&uevent, 0, 0, 0, true, &result);
    return result == ERR_OK;
}

bool wiimote_enable_ir_clock(uint8_t index, uint8_t enabled, uint8_t ack) {
    uint32_t result;
    uevent_t uevent;
    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_enable_ir_clock, 3, 0);

    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, enabled);
    uparam_set(u[3], u8, ack);

    Transx(&uevent, 0, 0, 0, true, &result);
    return result == ERR_OK;
}

bool wiimote_enable_audio_clock(uint8_t index, uint8_t enabled, uint8_t ack) {
    uint32_t result;
    uevent_t uevent;
    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_enable_audio_clock, 3, 0);

    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, enabled);
    uparam_set(u[3], u8, ack);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_read_status(uint8_t index) {
    uint32_t result;
    uevent_t uevent;

    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_read_status, 1, 0);

    uparam_set(u[1], u8, index);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_write_device(uint8_t index, uint8_t device, uint8_t mode, uint16_t offset, const uint8_t* src, uint8_t size) {
    uint32_t result;
    uevent_t uevent;

    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_write_device, 5, size);
    uparam_set(u[1], u8,  index);
    uparam_set(u[2], u8,  device);
    uparam_set(u[3], u8,  mode);
    uparam_set(u[4], u16, offset);
    uparam_set_tail_str(u, src, size);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_read_device(uint8_t index, uint8_t device, uint8_t mode, uint16_t offset, uint16_t size) {
    uint32_t result;
    uevent_t uevent;

    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_read_device, 5, 0);
    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, device);
    uparam_set(u[3], u8, mode);
    uparam_set(u[4], u16, offset);
    uparam_set(u[5], u16, size);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_write_audio(uint8_t index, const uint8_t* src, uint8_t size) {
    uint32_t result;
    uevent_t uevent;

    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_write_audio, 2, size);
    uparam_set(u[1], u8, index);
    uparam_set_tail_str(u, src, size);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_mute_audio(uint8_t index, uint8_t mute, uint8_t ack) {
    uint32_t result;
    uevent_t uevent;

    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_mute_audio, 3, 0);
    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, mute);
    uparam_set(u[3], u8, ack);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

bool wiimote_enable_ir(uint8_t index, uint8_t enabled, uint8_t ack) {
    uint32_t result;
    uevent_t uevent;

    uparam_t* u = SetupUevent(&uevent, umajor_hidhost, hidhost_minor_cmd_wiimote_enable_ir, 3, 0);
    uparam_set(u[1], u8, index);
    uparam_set(u[2], u8, enabled);
    uparam_set(u[3], u8, ack);

    Transx(&uevent, 0, 0, 0, true, &result);    
    return result == ERR_OK;
}

uint8_t get_command(uint8_t index, uint8_t* command)
{	
	uevent_t uevent;
	uint32_t result;
	uint8_t i;

	if (index == 0) { return 0; }

	uparam_t* u = SetupUevent(&uevent, umajor_loader, loader_minor_cmd_get_command, 1, 0);	
	uparam_set(u[1], u8, index);
	//Transx(&uevent, 0, 0, 0, true, &result);
	uparam_t* v = Transx(&uevent, umajor_loader, loader_minor_cfm_get_command, /*udata_usb_extra_cmd_uevent*/ 0 , false, &result, 1000);		

	if (result == ERR_OK)
	{	
		if (v && v[1].u.u8) {
			for (i = 0; i < v[1].u.u8; i++)
			{
				command[i] = v[i + 2].u.u8;
			}			

			return 1;
		}
		else {
			return 0;
		}
	}

	return 0;
}

int  connected_num(void){
    return _connected_num;
}

int  connected_first_index(void){
    return _connected_first;
}

///////////////////////////////////////////////////////////////////////////////
// WiiPC Internal Common Interface
///////////////////////////////////////////////////////////////////////////////
void  ReaderStart(void){
    Reader()->start();
}

void  ReaderStop(void){
    Reader()->stop();
}

void  ReaderReset(int __core index){
    Reader()->reset(index);
}

void  ReaderResetAll(void){
    Reader()->reset_all();
}

void  WriterStart(void){
    Writer()->start();
}

int   ReaderTryRead(DWORD timeout, udata_t* _udata){
    return Reader()->try_read(timeout, _udata);
}

void  WriterStop(void){
    Writer()->stop();
}

void  WriterResetAll(void){
    Writer()->reset_all();
}

void WriterReset(int __core index){
    Writer()->reset(index);
}

int   WriterWrite(udata_t* udata){
    return Writer()->write(udata);
}

bool  ManagerIsBridged(void){
    return Manager()->is_bridged();
}

void  ManagerNotify(udata_t* d, uevent_t* e){
    Manager()->notify(d, e);
}

void  ManagerUsbDisconnect(void){
    Manager()->usb_disconnect();
}

void  ManagerUsbConnect(void){
    Manager()->usb_connect();
}

bool  ManagerUsbTryConnect(void){
    return Manager()->usb_try_connect();
}

bool  ManagerIsUsbConnected(void){
    return Manager()->usb_handle != 0;
}

bool  ManagerIsUsbDisconnected(void){
    return Manager()->usb_handle == 0;
}

void* ManagerUsbHandle(void){
    return Manager()->usb_handle;
}

void  ManagerStart(void){
    Manager()->start();
}

void  PresenterStart(void){
    Presenter()->start();
}

void  PresenterStop(void){
    Presenter()->stop();
}

#if 0
void  MapperInit    (unsigned int _g_wiimote_sources[4], uint8_t* _conn_bitmap){
    Mapper()->init(_g_wiimote_sources, _conn_bitmap);
}

void  MapperReset   (void){
    Mapper()->reset();
}

void  MapperRemap   (void){
    Mapper()->remap();
}

void  MapperDump    (void){
    Mapper()->dump();
}
#endif

void  EventerStart  (void){
    Eventer()->start();
}

event_id_t  EventerSet   (event_callback_t callback, uint16_t m, void* p, uint32_t delay_ms, uint32_t group){
    return Eventer()->set_event(callback, m, p, delay_ms, group);
}

bool  __ok  EventerCancel(event_id_t event_id, uint16_t* pm, void** pp){
    return Eventer()->cancel_event(event_id, pm, pp);
}

uint32_t __num EventerCancelGroup(uint32_t group, event_callback_t cancel_callback){
    return Eventer()->cancel_group(group, cancel_callback);
}


bool __valid MapperWiiPCToCore(int __wiipc index, int* __core mapped_index){
    return Manager()->wiipc_to_core(index, mapped_index);
}

bool __valid MapperCoreToWiiPC(int __core index, int* __wiipc mapped_index){
    return Manager()->core_to_wiipc(index, mapped_index);
}

bool __valid ManagerIsWmConnected(int __wiipc index){
    return Manager()->is_wm_connected(index);
}


///////////////////////////////////////////////////////////////////////////////
//  Interface to Core / Wiimote
///////////////////////////////////////////////////////////////////////////////
void Initialize(void){
    //dprintf("Initialize ");
    EventerStart();
    ManagerStart();
    PresenterStart();
}

void Shutdown(void){	
	// Webb 
	// state = Bridged, I set all wiimote to basic.
	if (Manager()->state == Manager()->Bridged) {
		Manager()->set_all_wiimotes_to_basic();
	}
}


void RealShutdown(void){
    if(_Presenter) {
        _Presenter->stop();
    }

    if(_Manager){
        _Manager->stop();
    }

    if(_Reader){
        _Reader->stop();
    }

    if(_Writer){
        _Writer->stop();
    }

    if(_Manager) {
        delete _Manager;
        _Manager = 0;
    }

    if(_Reader) {
        delete _Reader;
        _Reader = 0;
    }

    if(_Writer) {
        delete _Writer;
        _Writer = 0;
    }
}

void InterruptChannel(int number, uint16_t channel, const void* buffer, uint32_t size){
    /* Core to WiiPC */

    #if 0
    dprintf("W[%d]", number);
    dprintbuf(buffer, size);
    dprintf("\n");
    #endif

    int r = Writer()->write_interrupt(number, channel, buffer, size);

    if(r != ERR_OK){
        //dprintf("writeNG=%d\n", r);
    }
}

void CB_wm_logical_disconnected(uint16_t m, void* p){
    if(Manager()->set_wiimote_to_basic(m)){
        wiimote_set_leds(m & 3, 2+4, 0);
    }
}

void CB_forced_reattach(uint16_t m, void* p){
    Manager()->attach_core_wiimote(m & 3);
}

void ControlChannel(int number, uint16_t channel, const void* buffer, uint32_t size){
    /* Core to WiiPC */
    int  i;

    //dprintf("ControlChannel number=%d channel=%d\n", number, channel);

    if(MapperCoreToWiiPC(number, &i)) {
        if(99 == channel) {
            /* disconnection */
            //EventerSet(CB_wm_logical_disconnected, i, 0, 100, 1 /* group */);
            //EventerSet(CB_forced_reattach, number, 0, 1000, 1 /* group */);
            Manager()->notify_event(umajor_manager, manager_minor_ind_core_wm_disc, number);
        }else{
            InterruptChannel(number, channel, buffer, size);
            uint8_t* p = (uint8_t*)buffer;

            if((p[0] & 0xF0) == 0x50) {
                /* SET_REPORT */
                uint8_t handshake_ok = 0;
                Wiimote::CoreWiimoteControlChannel(number, channel, &handshake_ok, sizeof(handshake_ok));
            }
        }
    }
}

void Update(int index){
    report_t r;	 

    if(Reader()->feed(&r, index)) {            		
        //Wiimote::CoreWiimoteInterruptChannel(index, current_connection, r.data);
        #if 1
        if (Wii_Balance_Board_Flag == 1) {
            if (index == 0) {                    
                Wiimote::CoreWiimoteInterruptChannel(index, Index_0_Header, r.data);
            }
            else if (index == 1) {                
                Wiimote::CoreWiimoteInterruptChannel(index, Index_1_Header, r.data);
            }
        }
        else {			
            Wiimote::CoreWiimoteInterruptChannel(index, current_connection, r.data);
        }
        #endif        
    }
    else{
        //dprintf("NoData ");        		
    }
}

} /* namespace WiiPC */

