#include "iofile.h"

#include "CommonTypes.h"
#include "CommonFuncs.h"
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>
#include <io.h>
#include <wchar.h>

void StripTailDirSlashes(std::wstring& fname){
	if (fname.length() > 1)
	{
		size_t i = fname.length() - 1;
		while (fname[i] == '/' || fname[i] == '\\'){
			fname[i--] = '\0';
        }
	}
	return;
}

bool Exists(const std::wstring& filename)
{
	struct stat64 file_info;

	std::wstring copy(filename);
	StripTailDirSlashes(copy);	
	int result = _wstati64(copy.c_str(), &file_info);

	return (result == 0);
}

// Returns true if filename is a directory
bool IsDirectory(const std::wstring& filename)
{
	struct stat64 file_info;

	std::wstring copy(filename);
	StripTailDirSlashes(copy);

	int result = _wstati64(copy.c_str(), &file_info);

	if (result < 0) {
		return false;
	}
	
	// C++ Builder is S_ISDIR(file_info.st_mode)
	return ((file_info.st_mode) & S_IFDIR);
}

u64 FileSize(const std::wstring& filename)
{
	if (!Exists(filename)){
		return 0;
	}

	if (IsDirectory(filename)){
		return 0;
	}

	struct stat64 buf;

	if(_wstati64(filename.c_str(), &buf) == 0) {
		return buf.st_size;
	}

	return 0;
}

u64 FileSize(FILE *f){
	// can't use off_t here because it can be 32-bit
	u64 pos = ftello(f);

	if (fseeko(f, 0, SEEK_END) != 0) {
		return 0;
	}

	u64 size = ftello(f);

	if ((size != pos) && (fseeko(f, pos, SEEK_SET) != 0)) {
		return 0;
	}
	return size;
}


IOFile::IOFile()
: m_file(NULL), m_good(true)
{}

IOFile::IOFile(std::FILE* file)
: m_file(file), m_good(true)
{}

#if 0
IOFile::IOFile(const std::string& filename, const char openmode[])
: m_file(NULL), m_good(true)
{
	Open(filename, openmode);
}
#endif

IOFile::IOFile(const std::wstring& filename, const char openmode[]){
	Open(filename, openmode);
}

IOFile::~IOFile()
{
    Close();
}

IOFile::IOFile(IOFile&& other)
: m_file(NULL), m_good(true)
{
    Swap(other);
}

IOFile& IOFile::operator=(IOFile&& other)
{
    IOFile((IOFile&&)other).Swap(*this);
    return *this;
}

void IOFile::Swap(IOFile& other)
{
    std::swap(m_file, other.m_file);
    std::swap(m_good, other.m_good);
}

#if 0
bool IOFile::Open(const std::string& filename, const char openmode[])
{
    Close();
#ifdef _WIN32
    _tfopen_s(&m_file, UTF8ToTStr(filename).c_str(), UTF8ToTStr(openmode).c_str());
#else
    m_file = fopen(filename.c_str(), openmode);
#endif

    m_good = IsOpen();
    return m_good;
}
#endif

bool IOFile::Open(const std::wstring& filename, const char openmode[]){
	Close();
	std::string m(openmode);
	std::wstring wopenmode;

	wopenmode.assign(m.begin(), m.end());
	_wfopen_s(&m_file, filename.c_str(), wopenmode.c_str());

    m_good = IsOpen();
    return m_good;
}

bool IOFile::Close()
{
    if (!IsOpen() || 0 != std::fclose(m_file))
        m_good = false;

    m_file = NULL;
    return m_good;
}

std::FILE* IOFile::ReleaseHandle()
{
    std::FILE* const ret = m_file;
    m_file = NULL;
    return ret;
}

void IOFile::SetHandle(std::FILE* file)
{
    Close();
    Clear();
    m_file = file;
}

u64 IOFile::GetSize()
{
    if (IsOpen())
        return FileSize(m_file);
    else
        return 0;
}

bool IOFile::Seek(s64 off, int origin)
{
    if (!IsOpen() || 0 != fseeko(m_file, off, origin))
        m_good = false;

    return m_good;
}

u64 IOFile::Tell()
{
    if (IsOpen())
        return ftello(m_file);
    else
        return -1;
}

bool IOFile::Flush()
{
    if (!IsOpen() || 0 != std::fflush(m_file))
        m_good = false;

    return m_good;
}

bool IOFile::Resize(u32 size){
	if (!IsOpen() || 0 != chsize(_fileno(m_file), size)){
		m_good = false;
	}

	return m_good;
}

