#include "DriveWbfsBlob.h"
#include "TsAny.hpp"
#include <windows.h>
#include <winioctl.h>
#include <stdlib.h>
#include <dconsole.h>

namespace DiscIO
{

static DriveWbfsReader MetaDriveWbfsReader(IBlobReaderTypeInfo("dwbfs", 0));

static inline unsigned short _byteswap_ushort(unsigned short value)
{
	return (value >> CHAR_BIT) | (value << CHAR_BIT);
}

static inline unsigned long _byteswap_ulong(unsigned long value)
{
	return _byteswap_ushort((unsigned short)value) << (2 * CHAR_BIT) |
		   _byteswap_ushort((unsigned short)(value >> (2 * CHAR_BIT)));
}

static inline u64 align(u64 value, u64 bounds){
    return (value + (bounds -1)) & (~(bounds - 1));
}

int DriveWbfsReader::GetDriveCapacity(const wchar_t* path, u32* sector_size, u32* sector_count, u64* length){
    DISK_GEOMETRY dg;
    PARTITION_INFORMATION pi;
    DWORD bytes;
		
	// webb
	// wchar_t* convert to a char*
	size_t origsize = wcslen(path) + 1;
    const size_t newsize = 100;
    size_t convertedChars = 0;
    char nstring[newsize];
    wcstombs_s(&convertedChars, nstring, origsize, path, _TRUNCATE);	
	
    HANDLE h = CreateFile(nstring, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);

    if(h == INVALID_HANDLE_VALUE) {
        return -1;
    }

    if(DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &dg, sizeof(DISK_GEOMETRY), &bytes, NULL) == NULL) {
        CloseHandle(h);
        return -2;
    }

    if(sector_size) {
        *sector_size = dg.BytesPerSector;
    }

    if(DeviceIoControl(h, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0, &pi, sizeof(PARTITION_INFORMATION), &bytes, NULL) == FALSE) {
        CloseHandle(h);
        return -3;
    }

    if(length) {
        *length = pi.PartitionLength.QuadPart;
    }

    if(sector_count) {
        *sector_count = (u32)(pi.PartitionLength.QuadPart / dg.BytesPerSector);
    }

    CloseHandle(h);

    return 0;
}

DriveWbfsReader::DriveWbfsReader(const wchar_t* drive_letter, int disc_index /* from_zero */)
    :handle(), m_disc_index(disc_index), is_disc_loaded()
{
	std::wstring d(drive_letter);
	if(d.length() > 2){
		d[2] = 0;
	}

    std::wstring path(L"\\\\.\\" + d);	
	
	// webb
	// wchar_t* convert to a char*
	size_t origsize = wcslen(path.c_str()) + 1;
    const size_t newsize = 100;
    size_t convertedChars = 0;
    char nstring[newsize];
    wcstombs_s(&convertedChars, nstring, origsize, path.c_str(), _TRUNCATE);	

    if(GetDriveCapacity(path.c_str(), &drv_sector_size, &drv_sector_count, &m_size)){
        return;
    }

    handle = CreateFile(nstring, 
                        GENERIC_READ, 
                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                        NULL, 
                        OPEN_EXISTING, 
                        FILE_FLAG_RANDOM_ACCESS, 
                        NULL);

    if(handle == INVALID_HANDLE_VALUE) {
        handle = 0;
        return;
	}

    /* read the first sector */
    RawSectorSeek(0);
    drv_sector.resize(drv_sector_size);

    RawSectorRead(1, drv_sector.get());
    drv_sector_index = 0;
	drv_sector_offset = 0;

    ReadHeader();
}

DriveWbfsReader* DriveWbfsReader::Create(const wchar_t* filename){
    /* 
        "G"        => at drive G, index = 0
        "G:"       => at drive G, index = 0
        "G:13"     => at drive G, index = 13
        "G:ASMB01" => at drive G, ID = "ASMB01"
    */
    wstring w(filename);
	wchar_t drive[3];
	int index = -1;

	if(w.length() >= 1) {
		drive[0] = w[0];
		drive[1] = L':';
        drive[2] = 0;
    }

    if(w.length() >= 2) {
        if(w.substr(1, 1) != L":") {
            return 0;
        }
    }

    if(w.length() > 2) {
        index = wstring_to_long_def(w.substr(2), -1);
    }

    DriveWbfsReader* instance = new DriveWbfsReader(drive, index);	

#if 1
    if(index < 0) {
        index = instance->SetDiscIndexByID(strwtos(w.substr(2, 6)));
	}

	instance->config = w.substr(2);

    instance->LoadDisc();
    return instance;
#endif
	
}

bool DriveWbfsReader::RawSectorSeek(u64 sector_num){
    if(!handle) return false;
    LARGE_INTEGER large;
    large.QuadPart = sector_num * drv_sector_size;
    return SetFilePointerEx(handle, large, NULL, FILE_BEGIN); 
}

bool DriveWbfsReader::RawSectorRead(u64 sector_count, u8* out_ptr){
	if(!handle) return false;
	DWORD dummy;
    return ReadFile(handle, out_ptr, sector_count * drv_sector_size, &dummy, NULL);
}

bool DriveWbfsReader::Seek(u64 offset){
    if(!handle) {
        return false;
    }

    u64 I = offset / drv_sector_size;
    u32 J = offset % drv_sector_size;

    if(I != drv_sector_index) {
        if(RawSectorSeek(I) && RawSectorRead(1, drv_sector.get())){
            drv_sector_index = I;
            drv_sector_offset = J;
            return true;
        }
	}

	drv_sector_offset = J;
    return true;
}

bool DriveWbfsReader::Skip(u64 nbytes){
    if(!handle) {
        return false;
    }
    u64 old = drv_sector_index;
    u64 I = drv_sector_index * drv_sector_size + drv_sector_offset + nbytes;

    drv_sector_index  = I / drv_sector_size;
    drv_sector_offset = I % drv_sector_size;

    if(old != drv_sector_index) {
        return RawSectorRead(1, drv_sector.get());
    }

    return true;
}

bool DriveWbfsReader::RawRead(u64 nbytes, u8* out_ptr){
    u32 n;
    u32 m;
    u8* p = drv_sector.get();

    while(nbytes) {
        n = drv_sector_size - drv_sector_offset;
        m = nbytes >= n ? n : nbytes;

        memcpy(out_ptr, &p[drv_sector_offset], m);

		nbytes -= m;
		drv_sector_offset += m;
        out_ptr += m;

        if(drv_sector_offset >= drv_sector_size) {
            if(!RawSectorRead(1, p)){
                return false;
            }
            drv_sector_index ++;
            drv_sector_offset = 0;
        }
    }

    return true;
}

bool DriveWbfsReader::RawRead(u64 offset, u64 nbytes, u8* out_ptr){
    return Seek(offset) && RawRead(nbytes, out_ptr);
}

u8   DriveWbfsReader::RawReadBigEndian8 (void){
    u8 v;

    RawRead(1, &v);
    return v;
}

u16  DriveWbfsReader::RawReadBigEndian16(void){
    u16 v;

	RawRead(2, (u8*)&v);
	return _byteswap_ushort(v);
}

u32  DriveWbfsReader::RawReadBigEndian32(void){
	u32 v;
	RawRead(4, (u8*)&v);
	return _byteswap_ulong(v);
}

bool DriveWbfsReader::ReadHeader(void){
    if(!Seek(4)) return false;
    
	hd_sector_count = RawReadBigEndian32();
    hd_sector_shift = RawReadBigEndian8();
    hd_sector_size = 1 << hd_sector_shift;

    wbfs_sector_shift = RawReadBigEndian8();
    wbfs_sector_size  = 1 << wbfs_sector_shift;
	wbfs_sector_count = m_size / wbfs_sector_size;

	if(wbfs_sector_size < wii_sector_size){
        return false;
	}

    m_blocks_per_disc = (wii_sector_count * wii_sector_size) / wbfs_sector_size;
    m_disc_info_size  = align(wii_disc_header_size + m_blocks_per_disc * 2, hd_sector_size);

    m_freeblks_lba = (wbfs_sector_size - wbfs_sector_count/8) >> hd_sector_shift;
    m_max_disc = (m_freeblks_lba - 1)/(m_disc_info_size >> hd_sector_shift);

    disc_table.resize(m_max_disc);

    Skip(2);
    RawRead(m_max_disc, disc_table.get());

    return true;
}

string DriveWbfsReader::DiscID(int index){
    u8 a[7] = {0};

    if(index == -1) {
        index = m_disc_index;
    }

	if(index < m_max_disc && disc_table[index]){
		if(RawRead(hd_sector_size + index*m_disc_info_size, 6, a)){
			return string((char*)a);
		}
	}

    return "";
}

u16  DriveWbfsReader::DiscCount(void){
    if(!handle) {
        return 0;
    }

    u16 count = 0;

    for(u16 i=0; i<500; ++i) {
        if(disc_table[i]) {
            count ++;
        }
    }

    return count;
}

int DriveWbfsReader::FindDiscID(const string& id){
    for(int i=0; i<m_max_disc; ++i) {
        if(id == DiscID(i)) {
            return i;
        }
    }

    return 0;
}

int DriveWbfsReader::SetDiscIndexByID(const string& id){
    m_disc_index = FindDiscID(id);
    return m_disc_index;
}


bool DriveWbfsReader::LoadDisc(void){
    if(m_disc_index < 0) {
        return false;
    }

    if(m_disc_index >= m_max_disc) {
        return false;
    }

    if(disc_table[m_disc_index] == 0){
        return false;
    }

    if(is_disc_loaded) {
        return false;
    }
		
    m_wlba_table.resize(m_blocks_per_disc);	
    if(RawRead(hd_sector_size + wii_disc_header_size + m_disc_index * m_disc_info_size,
               m_blocks_per_disc*sizeof(u16), 
               (u8*)m_wlba_table.get())){
        is_disc_loaded = true;
    }

    return is_disc_loaded;
}

bool DriveWbfsReader::Read(u64 offset, u64 nbytes, u8* out_ptr){
    if(is_disc_loaded) {
        u64 base = offset >> wbfs_sector_shift;

        if(base < m_blocks_per_disc) {
            u64 cluster_address = wbfs_sector_size * _byteswap_ushort(m_wlba_table[base]);
            u64 cluster_offset  = offset & (wbfs_sector_size - 1);
            u64 final_address = cluster_address + cluster_offset;

            return RawRead(final_address, nbytes, out_ptr);
        }
    }

    return false;
}

bool DriveWbfsReader::Identify(void){
    u8 magic[4] = {0};

    return handle && 
           RawRead(0, 4, magic) &&
           magic[0] == 'W'   && 
           magic[1] == 'B'   && 
           magic[2] == 'F'   && 
           magic[3] == 'S';
}
} // namespace
