#include "VolumeWii.h"
#include "aes.h"

static const unsigned char g_MasterKey[16] = {0xeb,0xe4,0x2a,0x22,0x5e,0x85,0x93,0xe4,0x48,0xd9,0xc5,0x45,0x73,0x81,0xaa,0xf7};
static const unsigned char g_MasterKeyK[16] = {0x63,0xb8,0x2b,0xb4,0xf4,0x61,0x4e,0x2e,0x13,0xf2,0xfe,0xfb,0xba,0x4c,0x9b,0x7e};

namespace DiscIO
{

static TuVolumeWii MetaVolumeWii(VolumeTypeInfo("wii", 0));

TuVolumeWii::TuVolumeWii(TsIBlobReader _blob)
    :
    TuVolume(_blob),
    m_buffer(0x8000),
    m_LastDecryptedBlock(0x8000),
    dataOffset(0x20000), 
	m_LastDecryptedBlockOffset(-1),
	volumeNum(-1),
	partitionGroup(),
	volumeType(),
	is_read_ready()
	{}


bool TuVolumeWii::Identify(void){
    SPartitionGroup PartitionGroup[4];
    bool Korean;

	if(!blob.get())		  return false;
	if(!blob->Identify()) return false;
	if(is_read_ready)	  return false;

    RawRead(0x3, 1, &region);
    Korean = (region == 'K');

    u32 WiiMagic          = RawReadBigEndian32(0x18);
    u32 WiiContainerMagic = RawReadBigEndian32(0x60);

    if(WiiMagic == 0x5D1C9EA3){
        if(WiiContainerMagic == 0) {
            discType = DISC_TYPE_WII_CONTAINER;
        }else{
            discType = DISC_TYPE_WII;
        }
    }else{
        discType = DISC_TYPE_NONE;
        return false;
    }

    numPartitions = RawReadBigEndian32(0x40000 + (partitionGroup * 8));
    partitionsOffset = (u64)RawReadBigEndian32(0x40000 + (partitionGroup * 8) + 4) << 2;

    // Check if we're looking for a valid partition
    if ((int)volumeNum != -1 && volumeNum > numPartitions){
        return false;
    }

    // read all partitions
    for (u32 x = 0; x < 4; x++){
        for (u32 i = 0; i < numPartitions; i++){
            SPartition Partition;
            Partition.Offset = ((u64)RawReadBigEndian32(partitionsOffset + (i * 8) + 0)) << 2;
            Partition.Type   = RawReadBigEndian32(partitionsOffset + (i * 8) + 4);
            PartitionGroup[x].PartitionsVec.push_back(Partition);
        }
    }

    // return the partition type specified or number
    // types: 0 = game, 1 = firmware update, 2 = channel installer
    //  some partitions on ssbb use the ascii title id of the demo VC game they hold...
    for (size_t i = 0; i < PartitionGroup[partitionGroup].PartitionsVec.size(); i++){
        const SPartition& rPartition = PartitionGroup[partitionGroup].PartitionsVec.at(i);

        if (rPartition.Type == volumeType || i == volumeNum){
            u8 SubKey[16];
            RawRead(rPartition.Offset + 0x1bf, 16, SubKey);

            u8 IV[16];
            memset(IV, 0, 16);
            RawRead(rPartition.Offset + 0x44c, 8, IV);

            AES_KEY AES_KEY;
            AES_set_decrypt_key((Korean ? g_MasterKeyK : g_MasterKey), 128, &AES_KEY);

            AES_cbc_encrypt(SubKey, volumeKey, 16, &AES_KEY, IV, AES_DECRYPT);

            // -1 means the caller just wanted the partition with matching type
            if ((int)volumeNum == -1 || i == volumeNum){
                //return new CVolumeWiiCrypted(&_rReader, rPartition.Offset, VolumeKey);
                volumeOffset = rPartition.Offset;
                AES_set_decrypt_key(volumeKey, 128, &m_AES_KEY);

                is_read_ready = true;
                return true;
            }
        }
    }

    return false;
}

bool TuVolumeWii::Read(u64 _ReadOffset, u64 _Length, u8* _Buffer){
    if(!is_read_ready) {
        return false;
    }

	u8* B = m_buffer.get();
	u8* L = m_LastDecryptedBlock.get();

    while (_Length > 0)
    {
        static unsigned char IV[16];

        // math block offset
        u64 Block  = _ReadOffset / 0x7C00;
        u64 Offset = _ReadOffset % 0x7C00;

        // read current block
        if (!RawRead(volumeOffset + dataOffset + Block * 0x8000, 0x8000, B)){
            return false;
        }

        if (m_LastDecryptedBlockOffset != Block){
            memcpy(IV, B + 0x3d0, 16);
            AES_cbc_encrypt(B + 0x400, L, 0x7C00, &m_AES_KEY, IV, AES_DECRYPT);

            m_LastDecryptedBlockOffset = Block;
        }

        // copy the encrypted data
        u64 MaxSizeToCopy = 0x7C00 - Offset;
        u64 CopySize = (_Length > MaxSizeToCopy) ? MaxSizeToCopy : _Length;
        memcpy(_Buffer, &L[(u32)Offset], (size_t)CopySize);

        // increase buffers
        _Length     -= CopySize;
        _Buffer	    += (u32)CopySize;
        _ReadOffset += CopySize;
    }

    return true;
}

string TuVolumeWii::GetID(void){
    char ID[7] = {0};
    return Read(0, 6, (u8*)ID) ? ID : "";
}

string TuVolumeWii::GetCountry(void){
    if(!is_read_ready) {
        return "";
    }

	switch (region) 
	{
		// Region free - fall through to European defaults for now
		case 'A':

		
		// PAL
		case 'D': // German
			return "COUNTRY_GERMANY";
			break;

		case 'X': // Used by a couple PAL games
		case 'Y': // German, french

		case 'L': // Japanese import to PAL regions
		case 'M': // Japanese import to PAL regions
		case 'S': // Spanish-speaking regions
		case 'P':
		case 'U': // Australia
			return "COUNTRY_EUROPE";
			break;
			
		case 'F':
			return "COUNTRY_FRANCE";
			break;
			
		case 'I':
			return "COUNTRY_ITALY";
			break;
			
		case 'R':
			return "COUNTRY_RUSSIA";
			break;

		// NTSC
		case 'E':
		case 'N': // Japanese import to USA and other NTSC regions
		case 'Z': // Prince Of Persia - The Forgotten Sands (WII)
			return "COUNTRY_USA";
			break;

		case 'J':
			return "COUNTRY_JAPAN";
			break;

		case 'K':
		case 'T': // Korea with English language
		case 'Q': // Korea with Japanese language
			return "COUNTRY_KOREA";
			break;			

		case 'O':
			return "COUNTRY_SDK";
			break;

		case 'W':
			return "COUNTRY_TAIWAN";
			break;
	}

	return "COUNTRY_UNKNOWN";
}

} // namespace